import os
import logging
import hashlib

logger = logging.getLogger(__name__)


def prepare_path(file_target):
    # as the path might not exist, we can not test if it is a dir beforehand
    # therefore, if it ends with a / it is considered a dir, otherwise, it is a regular file
    # and the following code works for both cases
    os.makedirs(os.path.abspath(file_target.rpartition('/')[0]), exist_ok=True)


# Function : md5sum
# Purpose : Get the md5 hash of a file stored in S3
# Returns : Returns the md5 hash that will match the ETag in S3
# https://stackoverflow.com/questions/6591047/etag-definition-changed-in-amazon-s3/28877788#28877788
# https://github.com/boto/boto3/blob/0cc6042615fd44c6822bd5be5a4019d0901e5dd2/boto3/s3/transfer.py#L169
def md5s3(file_like,
          multipart_threshold=8 * 1024 * 1024,
          multipart_chunksize=8 * 1024 * 1024):
    md5hash = hashlib.md5()
    file_like.seek(0)
    filesize = 0
    block_count = 0
    md5string = b''
    for block in iter(lambda: file_like.read(multipart_chunksize), b''):
        md5hash = hashlib.md5()
        md5hash.update(block)
        md5string += md5hash.digest()
        filesize += len(block)
        block_count += 1

    if filesize > multipart_threshold:
        md5hash = hashlib.md5()
        md5hash.update(md5string)
        md5hash = md5hash.hexdigest() + "-" + str(block_count)
    else:
        md5hash = md5hash.hexdigest()

    file_like.seek(0)
    # https://github.com/aws/aws-sdk-net/issues/815
    return '"{}"'.format(md5hash)



from click.exceptions import UsageError

from . import config


class EnvfilePathNotDefinedUsageError(UsageError):
    def __init__(self, message=None, *args, **kwargs):
        message = message or ''
        if message:
            message += '\n'
        super().__init__(message, *args, **kwargs)

        error_msg = 'Set the environemnt variable S3CONF or provide a section from an existing config file.'
        try:
            sections_detected = ''
            for section in config.ConfigFileResolver(config.LOCAL_CONFIG_FILE).sections():
                sections_detected += '    {}\n'.format(section)
        except FileNotFoundError:
            pass
        if sections_detected:
            sections_detected = '\n\nThe following sections were detected:\n\n' + sections_detected

        self.message += error_msg + sections_detected


class LocalCopyOutdated(UsageError):
    def __init__(self, message=None, *args, ctx=None):
        message = message or ''
        if message:
            message = message % args
            message += '\n'
        super().__init__(message, ctx=ctx)


class EnvfilePathNotDefinedError(Exception):
    pass


class FileDoesNotExist(FileNotFoundError):
    pass




import os
import sys
from .client import main as _main


def main():
    # fixing the name of the entrypoint when the module is executed as
    # python -m s3conf
    # https://docs.python.org/3/library/__main__.html
    entrypoint_name = os.path.basename(sys.argv[0])
    if entrypoint_name == '__main__.py':
        entrypoint_name = 's3conf'

    _main(prog_name=entrypoint_name)


if __name__ == '__main__':
    main()



import os
import logging
import shlex

import editor
import click
from editor import get_editor, get_editor_args
from click_log import core


# apply patches that allow editor with args
# https://github.com/fmoo/python-editor/pull/15
def _get_editor():
    executable = get_editor()
    return shlex.split(executable)[0]


def _get_editor_args(editor):
    args = get_editor_args(editor)
    editor = os.environ.get('VISUAL') or os.environ.get('EDITOR')
    if editor:
        args = shlex.split(editor)[1:] + args
    return args


editor.get_editor = _get_editor
editor.get_editor_args = _get_editor_args


# Creating our own handler class that always uses stderr to output logs.
# This way, we can avoid mixing logging information with actual output from
# the command line client.
class MyClickHandler(logging.Handler):
    def emit(self, record):
        try:
            msg = self.format(record)
            click.echo(msg, err=True)
        except (KeyboardInterrupt, SystemExit):
            raise
        except Exception:
            self.handleError(record)


core._default_handler = MyClickHandler()
core._default_handler.formatter = core.ColorFormatter()

# adding color to INFO log messages as well
core.ColorFormatter.colors['info'] = dict(fg='green')



import os
import codecs
import logging
import json
from shutil import rmtree

from .utils import prepare_path, md5s3
from . import exceptions, files, storages, config

logger = logging.getLogger(__name__)
__escape_decoder = codecs.getdecoder('unicode_escape')


def parse_dotenv(data):
    for line in data.splitlines():
        line = line.strip()
        if line and not line.startswith('#') and '=' in line:
            k, _, v = line.partition('=')

            # Remove any leading and trailing spaces in key, value
            k, v = k.strip(), v.strip().encode('unicode-escape').decode('ascii')

            if v and v[0] == v[-1] in ['"', "'"]:
                v = __escape_decoder(v[1:-1])[0]

            yield k, v


def unpack_list(files_list):
    files_pairs = files_list.split(';') if files_list else []
    files_map = []
    for file_map in files_pairs:
        file_source, _, file_target = file_map.rpartition(':')
        if file_source and file_target:
            files_map.append((file_source, file_target))
    return files_map


def phusion_dump(environment, path):
    prepare_path(path if path.endswith('/') else path + '/')
    for k, v in environment.items():
        with open(os.path.join(path, k), 'w') as f:
            f.write(v + '\n')


def change_root_dir(file_path, root_dir=None):
    if root_dir:
        file_path = os.path.join(root_dir, file_path.lstrip('/'))
    return file_path


def expand_path(path, path_target):
    mapping = []
    if os.path.isdir(path):
        for root, dirs, files in os.walk(path):
            for file in files:
                file_source = os.path.join(root, file)
                file_target = os.path.join(path_target,
                                           storages.strip_prefix(os.path.join(root, file), path).lstrip('/'))
                mapping.append((file_source, file_target))
    else:
        mapping.append((path, path_target))
    return mapping


def raise_out_of_sync(local_file, remote_file):
    raise exceptions.LocalCopyOutdated(
        'Upsync failed, target file probably changed since last downsync. Run downsync and redo your '
        'modifications to avoid conflicts. Offending file:\n\n    %s -> %s ',
        local_file,
        remote_file
    )


class S3Conf:
    def __init__(self, storage=None, settings=None):
        self.settings = settings or config.Settings()
        self.storage = storage or storages.S3Storage(settings=self.settings)

    @property
    def environment_file_path(self):
        # resolving environment file path
        file_name = self.settings.get('S3CONF')
        if not file_name:
            logger.error('Environemnt file name is not defined or is empty.')
            raise exceptions.EnvfilePathNotDefinedError()
        return file_name

    def upsync(self, local_root, map_files=False, force=False):
        # running operations
        local_environment = change_root_dir(os.path.basename(self.environment_file_path).lstrip('/'), local_root)

        if not force:
            md5_hash_file_name = os.path.join(local_root, '.md5')
            hashes = json.load(open(md5_hash_file_name))

        # checking if md5 hashes have not changed in remote storage since our last downsync
        # if force is set, ignore the hash check and upsync anyway
        if not force and hashes[local_environment] != self.get_envfile().md5():
            raise raise_out_of_sync(local_environment, self.environment_file_path)
        if map_files:
            local_mapping_root = os.path.join(local_root, 'root')
            env_vars = files.EnvFile(local_environment).as_dict()
            file_map_list = unpack_list(env_vars.get('S3CONF_MAP'))
            if not force:
                for remote_path, local_path in file_map_list:
                    local_path = change_root_dir(local_path, local_mapping_root)
                    mapping = expand_path(local_path, remote_path)
                    for local_file, remote_file in mapping:
                        if hashes[local_file] != self.storage.open(remote_file).md5():
                            raise raise_out_of_sync(local_file, remote_file)

        self.upload(local_environment, self.environment_file_path)
        if map_files:
            self.upload_mapping(file_map_list, root_dir=local_mapping_root)

    def downsync(self, local_root, map_files=False, wipe=False):
        if wipe:
            rmtree(local_root, ignore_errors=True)
        # running operations
        hashes = {}
        local_path = change_root_dir(os.path.basename(self.environment_file_path).lstrip('/'), local_root)
        hashes.update(self.download(self.environment_file_path, local_path))

        if map_files:
            env_vars = files.EnvFile(local_path).as_dict()
            local_mapping_root = os.path.join(local_root, 'root')
            if env_vars.get('S3CONF_MAP'):
                hashes.update(self.download_mapping(env_vars.get('S3CONF_MAP'), root_dir=local_mapping_root))

        md5_hash_file_name = os.path.join(local_root, '.md5')
        json.dump(hashes, open(md5_hash_file_name, 'w'), indent=4)
        return hashes

    def download_mapping(self, files, root_dir=None):
        if isinstance(files, str):
            files = unpack_list(files)
        hashes = {}
        for remote_file, local_file in files:
            hashes.update(self.download(remote_file, change_root_dir(local_file, root_dir)))
        return hashes

    def upload_mapping(self, files, root_dir=None):
        if isinstance(files, str):
            files = unpack_list(files)
        for remote_file, local_file in files:
            self.upload(change_root_dir(local_file, root_dir), remote_file)

    def download(self, path, path_target, force=False):
        hashes = {}
        logger.info('Downloading %s to %s', path, path_target)
        for md5hash, file_path in self.storage.list(path):
            if path.endswith('/') or not path:
                target_name = os.path.join(path_target, file_path)
            else:
                target_name = path_target
            prepare_path(target_name)
            target_file = files.File(target_name)
            existing_md5 = target_file.md5() if target_file.exists() and not force else None
            if not existing_md5 or existing_md5 != md5hash:
                source_name = os.path.join(path, file_path).rstrip('/')
                logger.debug('Transferring file %s to %s', source_name, target_name)
                with open(target_name, 'wb') as f:
                    # join might add a trailing slash, but we know it is a file, so we remove it
                    self.storage.open(source_name).read_into_stream(f)
            hashes[target_name] = md5hash
        return hashes

    def upload(self, path, path_target):
        logger.info('Uploading %s to %s', path, path_target)
        mapping = expand_path(path, path_target)
        for file_source, file_target in mapping:
            self.storage.write(open(file_source, 'rb'), file_target)

    def get_envfile(self):
        logger.info('Loading configs from {}'.format(self.environment_file_path))
        return files.EnvFile.from_file(self.storage.open(self.environment_file_path))

    def edit(self, create=False):
        files.EnvFile.from_file(self.storage.open(self.environment_file_path)).edit(create=create)



import logging
import io
import codecs
from tempfile import NamedTemporaryFile

import editor

from . import exceptions

logger = logging.getLogger(__name__)
__escape_decoder = codecs.getdecoder('unicode_escape')


def parse_env_var(value):
    """
    Split a env var text like

    ENV_VAR_NAME=env_var_value

    into a tuple ('ENV_VAR_NAME', 'env_var_value')
    """
    k, _, v = value.partition('=')

    # Remove any leading and trailing spaces in key, value
    k, v = k.strip(), v.strip().encode('unicode-escape').decode('ascii')

    if v and v[0] == v[-1] in ['"', "'"]:
        v = __escape_decoder(v[1:-1])[0]
    return k, v


class File:
    def __init__(self, name, storage=None):
        self.name = name
        self._storage = storage
        self._stream = None

    @property
    def storage(self):
        if not self._storage:
            from .storages import LocalStorage
            self._storage = LocalStorage()
        return self._storage

    def read_into_stream(self, stream):
        self.storage.read_into_stream(self.name, stream=stream)

    def read(self):
        return self.storage.read_into_stream(self.name).read()

    def exists(self):
        if list(self.storage.list(self.name)):
            return True
        return False

    def md5(self):
        md5hash, _ = next(self.storage.list(self.name))
        return md5hash

    def write(self, data):
        if isinstance(data, str):
            data = data.encode('utf-8')
        self.storage.write(io.BytesIO(data), self.name)

    def edit(self, create=False):
        with NamedTemporaryFile(mode='rb+', buffering=0) as f:
            original_data = b''
            try:
                original_data = self.read()
            except exceptions.FileDoesNotExist:
                if not create:
                    raise
            f.write(original_data)
            edited_data = editor.edit(filename=f.name)

        if edited_data != original_data:
            self.write(edited_data)
        else:
            logger.warning('File not changed. Nothing to write.')


class EnvFile(File):
    @classmethod
    def from_file(cls, obj):
        obj.__class__ = cls
        return obj

    def as_dict(self):
        env_dict = {}
        try:
            lines = str(self.read(), 'utf-8').splitlines()
            lines = [line for line in lines if line and not line.startswith('#') and '=' in line]
            env_dict = dict(parse_env_var(line) for line in lines)
        except FileNotFoundError:
            pass
        return env_dict

    def set(self, value, create=False):
        new_key, new_value = parse_env_var(value)
        new_lines = []
        value_set = False
        try:
            for line in str(self.read(), 'utf-8').splitlines():
                key, value = parse_env_var(line)
                if key == new_key:
                    new_lines.append('{}={}'.format(new_key, new_value))
                    value_set = True
                else:
                    new_lines.append(line)
        except FileNotFoundError:
            if not create:
                raise
        if not value_set:
            new_lines.append('{}={}'.format(new_key, new_value))
        self.write('\n'.join(new_lines))

    def unset(self, unset_key):
        new_lines = []
        unset_done = False
        try:
            for line in str(self.read(), 'utf-8').splitlines():
                key, value = parse_env_var(line)
                if key == unset_key:
                    unset_done = True
                    continue
                else:
                    new_lines.append(line)
        except FileNotFoundError:
            logger.warning('File does not exist')

        if unset_done:
            self.write('\n'.join(new_lines))
        else:
            logger.info('Key %s not found in environemnt file, doing nothing...', unset_key)

    def from_dict(self, env_vars):
        file_data = ''
        for var_name, var_value in env_vars.items():
            file_data += '{}={}\n'.format(var_name, var_value)
        self.write(file_data)

    def update(self, another_file):
        env_vars = self.as_dict()
        env_vars.update(another_file.as_dict())
        self.from_dict(env_vars)



import os
import re
import subprocess

version_re = re.compile('^Version: (.+)$', re.M)


def get_version(repo_path=None):
    repo_path = os.path.abspath(repo_path or os.path.dirname(os.path.dirname(__file__)))
    if os.path.isdir(os.path.join(repo_path, '.git')):
        # Get the version using "git describe".
        cmd = 'git describe --tags --match [0-9]*'.split()
        try:
            version = subprocess.check_output(cmd, cwd=repo_path).decode().strip()
        except subprocess.CalledProcessError:
            print('Unable to get version number from git tags')
            exit(1)

        # Don't declare a version "dirty" merely because a time stamp has
        # changed. If it is dirty, append a ".dev" suffix to indicate a
        # development revision after the release.
        with open(os.devnull, 'w') as fd_devnull:
            subprocess.call(['git', 'status'],
                            stdout=fd_devnull, stderr=fd_devnull)

        cmd = 'git diff-index --name-only HEAD'.split()
        try:
            dirty = subprocess.check_output(cmd, cwd=repo_path).decode().strip()
        except subprocess.CalledProcessError:
            print('Unable to get git index status')
            exit(1)

        if dirty != '':
            version += '.dev'
    else:
        try:
            # Extract the version from the PKG-INFO file.
            with open(os.path.join(repo_path, 'PKG-INFO')) as f:
                version = version_re.search(f.read()).group(1)
        except FileNotFoundError:
            import pkg_resources
            version = pkg_resources.require('s3conf')[0].version
    return version



import os
import logging
import boto3
from botocore.exceptions import ClientError
from io import BytesIO

from .utils import prepare_path, md5s3
from .config import Settings
from .files import File
from . import exceptions


logger = logging.getLogger(__name__)


def strip_prefix(text, prefix):
    return text[len(prefix):] if text.startswith(prefix) else text


def strip_s3_path(path):
    bucket, _, path = strip_prefix(path, 's3://').partition('/')
    return bucket, path


class BaseStorage:
    def __init__(self, settings=None):
        self._settings = settings or Settings()

    def read_into_stream(self, file_path, stream=None):
        raise NotImplementedError()

    def open(self, file_name):
        logger.debug('Reading from %s', file_name)
        return File(file_name, storage=self)

    def write(self, f, file_name):
        raise NotImplementedError()

    def list(self, path):
        raise NotImplementedError()


class S3Storage(BaseStorage):
    def __init__(self, settings=None):
        super(__class__, self).__init__(settings=settings)
        self._resource = None

    @property
    def s3(self):
        logger.debug('Getting S3 resource')
        # See how boto resolve credentials in
        # http://boto3.readthedocs.io/en/latest/guide/configuration.html#guide-configuration
        if not self._resource:
            logger.debug('Resource does not exist, creating a new one...')
            self._resource = boto3.resource(
                's3',
                aws_access_key_id=self._settings.get('S3CONF_ACCESS_KEY_ID') or self._settings.get('AWS_ACCESS_KEY_ID'),
                aws_secret_access_key=self._settings.get('S3CONF_SECRET_ACCESS_KEY') or self._settings.get('AWS_SECRET_ACCESS_KEY'),
                aws_session_token=self._settings.get('S3CONF_SESSION_TOKEN') or self._settings.get('AWS_SESSION_TOKEN'),
                region_name=self._settings.get('S3CONF_S3_REGION_NAME') or self._settings.get('AWS_S3_REGION_NAME'),
                use_ssl=self._settings.get('S3CONF_S3_USE_SSL') or self._settings.get('AWS_S3_USE_SSL', True),
                endpoint_url=self._settings.get('S3CONF_S3_ENDPOINT_URL') or self._settings.get('AWS_S3_ENDPOINT_URL'),
            )
        return self._resource

    def read_into_stream(self, file_path, stream=None):
        try:
            bucket_name, file_name = strip_s3_path(file_path)
            stream = stream or BytesIO()
            bucket = self.s3.Bucket(bucket_name)
            bucket.download_fileobj(file_name, stream)
            stream.seek(0)
            return stream
        except ClientError as e:
            if e.response['Error']['Code'] == '404':
                logger.debug('File %s in bucket %s does not exist', file_name, bucket)
                raise exceptions.FileDoesNotExist('s3://{}/{}'.format(bucket_name, file_name))
            else:
                raise

    def write(self, f, file_name):
        logger.debug('Writing to %s', file_name)
        bucket, path_target = strip_s3_path(file_name)
        try:
            bucket = self.s3.create_bucket(Bucket=bucket)
        except ClientError as e:
            if e.response['Error']['Code'] == 'BucketAlreadyExists':
                bucket = self.s3.Bucket(bucket)
        bucket.upload_fileobj(f, path_target)

    def list(self, path):
        logger.debug('Listing %s', path)
        bucket_name, path = strip_s3_path(path)
        bucket = self.s3.Bucket(bucket_name)
        try:
            for obj in bucket.objects.filter(Prefix=path):
                if not obj.key.endswith('/'):
                    yield obj.e_tag, strip_prefix(obj.key, path)
        except ClientError as e:
            if e.response['Error']['Code'] == 'NoSuchBucket':
                logger.warning('Bucket does not exist, list() returning empty.')
            else:
                raise


class LocalStorage(BaseStorage):
    def _validate_path(self, path):
        if path.startswith('s3://'):
            raise ValueError('LocalStorage can not process S3 paths.')

    def read_into_stream(self, file_name, stream=None):
        try:
            self._validate_path(file_name)
            stream = stream or BytesIO()
            with open(file_name, 'rb') as f:
                stream.write(f.read())
            stream.seek(0)
            return stream
        except FileNotFoundError:
            raise exceptions.FileDoesNotExist(file_name)

    def write(self, f, file_name):
        self._validate_path(file_name)
        prepare_path(file_name)
        open(file_name, 'wb').write(f.read())

    def list(self, path):
        self._validate_path(path)
        if os.path.isdir(path):
            for root, dirs, files in os.walk(path):
                for file in files:
                    yield md5s3(open(file, 'rb')), strip_prefix(os.path.join(root, file), path)
        else:
            # only yields if it exists
            if os.path.exists(path):
                # the relative path of a file to itself is empty
                # same behavior as in boto3
                yield md5s3(open(path, 'rb')), ''



import logging
import subprocess
import shlex
import os
from shutil import rmtree

import click
from click.exceptions import UsageError
import click_log

from . import s3conf, config, files, exceptions, storages, __version__


logger = logging.getLogger(__name__)


STORAGES = {
    's3': storages.S3Storage,
    'local': storages.LocalStorage,
}


class SectionArgument(click.Argument):
    def handle_parse_result(self, *args, **kwargs):
        try:
            return super().handle_parse_result(*args, **kwargs)
        except click.exceptions.MissingParameter:
            raise exceptions.EnvfilePathNotDefinedUsageError()


@click.group(invoke_without_command=True)
@click.version_option(version=__version__)
@click.option('--edit', '-e', is_flag=True)
@click.option('--create',
              '-c',
              is_flag=True,
              help='When trying to edit a file, create it if it does not exist.')
@click.pass_context
# this sets the log level for this app only
@click_log.simple_verbosity_option('s3conf')
def main(ctx, edit, create):
    """
    Simple command line tool to help manage environment variables stored in a S3-like system. Facilitates editing text
    files remotely stored, as well as downloading and uploading files.
    """
    # configs this module logger to behave properly
    # logger messages will go to stderr (check __init__.py/patch.py)
    # client output should be generated with click.echo() to go to stdout
    try:
        click_log.basic_config('s3conf')
        logger.debug('Running main entrypoint')
        if edit:
            if ctx.invoked_subcommand is None:
                logger.debug('Using config file %s', config.LOCAL_CONFIG_FILE)
                config.ConfigFileResolver(config.LOCAL_CONFIG_FILE).edit(create=create)
                return
            else:
                raise UsageError('Edit should not be called with a subcommand.')
        # manually call help in case no relevant settings were defined
        if ctx.invoked_subcommand is None:
            click.echo(main.get_help(ctx))
    except exceptions.FileDoesNotExist as e:
        raise UsageError('The file {} does not exist. Try "-c" option if you want to create it.'.format(str(e)))


@main.command('env')
@click.argument('section',
                required=False)
@click.option('--map-files',
              '-m',
              is_flag=True,
              help='If defined, tries to map files from the storage to the local drive as defined by '
                   'the variable S3CONF_MAP read from the S3CONF file.')
@click.option('--phusion',
              is_flag=True,
              help='If set, dumps variables to --dump-path in for format used by the phusion docker image. '
                   'More information in https://github.com/phusion/baseimage-docker.')
@click.option('--phusion-path',
              default='/etc/container_environment',
              show_default=True,
              help='Path where to dump variables as in the phusion docker image format.')
@click.option('--quiet',
              '-q',
              is_flag=True,
              help='Do not print any environment variables output.')
@click.option('--edit',
              '-e',
              is_flag=True)
@click.option('--create',
              '-c',
              is_flag=True,
              help='When trying to edit a file, create it if it does not exist.')
def env(section, map_files, phusion, phusion_path, quiet, edit, create):
    """
    Reads the file defined by the S3CONF variable and output its contents to stdout. Logs are printed to stderr.
    See options for added functionality: editing file, mapping files, dumping in the phusion-baseimage format, etc.
    """
    try:
        logger.debug('Running env command')
        settings = config.Settings(section=section)
        storage = STORAGES['s3'](settings=settings)
        conf = s3conf.S3Conf(storage=storage, settings=settings)

        if edit:
            conf.edit(create=create)
        else:
            env_vars = conf.get_envfile().as_dict()
            if env_vars.get('S3CONF_MAP') and map_files:
                conf.download_mapping(env_vars.get('S3CONF_MAP'))
            if not quiet:
                for var_name, var_value in sorted(env_vars.items(), key=lambda x: x[0]):
                    click.echo('{}={}'.format(var_name, var_value))
            if phusion:
                s3conf.phusion_dump(env_vars, phusion_path)
    except exceptions.EnvfilePathNotDefinedError:
        raise exceptions.EnvfilePathNotDefinedUsageError()
    except exceptions.FileDoesNotExist as e:
        raise UsageError('The file {} does not exist. Try "-c" option if you want to create it.'.format(str(e)))


@main.command('exec')
@click.option('--map-files',
              '-m',
              is_flag=True,
              help='If defined, tries to map files from the storage to the local drive as defined by '
                   'the variable S3CONF_MAP read from the S3CONF file.')
@click.argument('section',
                required=False)
# to use option-like arguments use "--"
# http://click.pocoo.org/6/arguments/#option-like-arguments
@click.argument('command',
                required=False,
                nargs=-1,
                type=click.UNPROCESSED)
@click.pass_context
def exec_command(ctx, section, command, map_files):
    """
    Sets the process environemnt and executes the [COMMAND] in the same context. Does not modify the current shell
    environment.

    If the [COMMAND] has option-like arguments, use the standard POSIX pattern "--" to separate options
    from arguments. Considering our configuration in the "dev" section, we could write:

    s3conf -v info exec dev -- ping -v google.com
    """
    try:
        logger.debug('Running exec command')
        existing_sections = config.ConfigFileResolver(config.LOCAL_CONFIG_FILE).sections()
        command = ' '.join(command)
        if section not in existing_sections:
            command = '{} {}'.format(section, command) if command else section
            section = None

        if not command:
            logger.warning('No command detected.')
            click.echo(exec_command.get_help(ctx))
            return

        settings = config.Settings(section=section)
        storage = STORAGES['s3'](settings=settings)
        conf = s3conf.S3Conf(storage=storage, settings=settings)

        env_vars = conf.get_envfile().as_dict()
        if env_vars.get('S3CONF_MAP') and map_files:
            conf.download_mapping(env_vars.get('S3CONF_MAP'))

        current_env = os.environ.copy()
        current_env.update(env_vars)
        logger.debug('Executing command "%s"', command)
        subprocess.run(shlex.split(command), env=current_env, check=True)
    except exceptions.EnvfilePathNotDefinedError:
        raise exceptions.EnvfilePathNotDefinedUsageError()


@main.command('download')
@click.argument('remote_path')
@click.argument('local_path')
def download(remote_path, local_path):
    """
    Download a file or folder from the S3-like service.

    If REMOTE_PATH has a trailing slash it is considered to be a folder, e.g.: "s3://my-bucket/my-folder/". In this
    case, LOCAL_PATH must be a folder as well. The files and subfolder structure in REMOTE_PATH are copied to
    LOCAL_PATH.

    If REMOTE_PATH does not have a trailing slash, it is considered to be a file, and LOCAL_PATH should be a file as
    well.
    """
    storage = STORAGES['s3']()
    conf = s3conf.S3Conf(storage=storage)
    conf.download(remote_path, local_path)


@main.command('upload')
@click.argument('local_path')
@click.argument('remote_path')
def upload(remote_path, local_path):
    """
    Upload a file or folder to the S3-like service.

    If LOCAL_PATH is a folder, the files and subfolder structure in LOCAL_PATH are copied to REMOTE_PATH.

    If LOCAL_PATH is a file, the REMOTE_PATH file is created with the same contents.
    """
    storage = STORAGES['s3']()
    conf = s3conf.S3Conf(storage=storage)
    conf.upload(local_path, remote_path)


@main.command('downsync')
@click.argument('section', cls=SectionArgument)
@click.option('--map-files',
              '-m',
              is_flag=True,
              help='If defined, tries to map files from the storage to the local drive as defined by '
                   'the variable S3CONF_MAP read from the S3CONF file using the config folder as the '
                   'root directory.')
def downsync(section, map_files):
    """
    For each section defined in the local config file, creates a folder inside the local config folder
    named after the section. Downloads the environemnt file defined by the S3CONF variable for this section
    to this folder.
    """
    try:
        storage = STORAGES['s3']()
        settings = config.Settings(section=section)
        conf = s3conf.S3Conf(storage=storage, settings=settings)
        local_root = os.path.join(config.LOCAL_CONFIG_FOLDER, section)
        conf.downsync(local_root, map_files=map_files)
    except exceptions.EnvfilePathNotDefinedError:
        raise exceptions.EnvfilePathNotDefinedUsageError()


@main.command('upsync')
@click.argument('section', cls=SectionArgument)
@click.option('--map-files',
              '-m',
              is_flag=True,
              help='If defined, tries to map files from the storage to the local drive as defined by '
                   'the variable S3CONF_MAP read from the S3CONF file using the config folder as the '
                   'root directory.')
def upsync(section, map_files):
    """
    For each section defined in the local config file, look up for a folder inside the local config folder
    named after the section. Uploads the environemnt file named as in the S3CONF variable for this section
    to the remote S3CONF path.
    """
    try:
        storage = STORAGES['s3']()
        settings = config.Settings(section=section)
        conf = s3conf.S3Conf(storage=storage, settings=settings)
        local_root = os.path.join(config.LOCAL_CONFIG_FOLDER, section)
        conf.upsync(local_root, map_files=map_files)
    except exceptions.EnvfilePathNotDefinedError:
        raise exceptions.EnvfilePathNotDefinedUsageError()


@main.command('set')
@click.argument('section')
@click.argument('value',
                required=False)
@click.option('--create',
              '-c',
              is_flag=True,
              help='When trying to set a variable, create the file if it does not exist.')
def set(section, value, create):
    """
    Set value of a variable in an environment file for the given section.
    If the variable is already defined, its value is replaced, otherwise, it is added to the end of the file.
    The value is given as "ENV_VAR_NAME=env_var_value", e.g.:

    s3conf set test ENV_VAR_NAME=env_var_value
    """
    if not value:
        value = section
        section = None
    try:
        logger.debug('Running env command')
        settings = config.Settings(section=section)
        conf = s3conf.S3Conf(settings=settings)

        env_vars = conf.get_envfile()
        env_vars.set(value, create=create)
    except exceptions.EnvfilePathNotDefinedError:
        raise exceptions.EnvfilePathNotDefinedUsageError()


@main.command('unset')
@click.argument('section')
@click.argument('value',
                required=False)
def unset(section, value):
    """
    Unset a variable in an environment file for the given section.
    The value is given is the variable name, e.g.:

    s3conf unset test ENV_VAR_NAME
    """
    if not value:
        value = section
        section = None
    try:
        logger.debug('Running env command')
        settings = config.Settings(section=section)
        conf = s3conf.S3Conf(settings=settings)

        env_vars = conf.get_envfile()
        env_vars.unset(value)
    except exceptions.EnvfilePathNotDefinedError:
        raise exceptions.EnvfilePathNotDefinedUsageError()



import os
import logging
from configparser import ConfigParser

from .files import File


logger = logging.getLogger(__name__)


def _lookup_config_folder(initial_folder='.'):
    if not initial_folder:
        config_folder = os.path.join('.', '.s3conf')
        logger.debug('Config folder detected: %s', config_folder)
        return config_folder
    current_path = os.path.abspath(initial_folder)
    path_items = set(os.listdir(current_path))
    if '.s3conf' in path_items:
        s3conf_folder = os.path.join(current_path, '.s3conf')
        if os.path.isdir(s3conf_folder):
            config_folder = os.path.join(current_path, '.s3conf')
            logger.debug('Config folder detected: %s', config_folder)
            return config_folder
    return _lookup_config_folder(os.path.dirname(current_path) if current_path != '/' else None)


LOCAL_CONFIG_FOLDER = _lookup_config_folder()
LOCAL_CONFIG_FILE = os.path.join(LOCAL_CONFIG_FOLDER, 'config')


class EnvironmentResolver:
    def get(self, item, default=None):
        return os.environ.get(item, default)


class ConfigFileResolver:
    def __init__(self, config_file, section=None):
        self.config_file = os.path.expanduser(config_file)
        self.section = section or 'DEFAULT'
        self._config = None

    @property
    def config(self):
        if not self._config:
            self._config = ConfigParser()
            self._config.read(self.config_file)
        return self._config

    @config.setter
    def config(self, value):
        self._config = value

    def get(self, item, default=None, section=None):
        return self.config.get(section or self.section, item, fallback=default)

    def edit(self, create=False):
        File(self.config_file).edit(create=create)

    def sections(self):
        return self.config.sections()


class Settings:
    def __init__(self, section=None, config_file=None):
        config_file = config_file or LOCAL_CONFIG_FILE
        if section:
            self.resolvers = [
                ConfigFileResolver(config_file, section),
                EnvironmentResolver(),
            ]
        else:
            self.resolvers = [
                EnvironmentResolver(),
                ConfigFileResolver(config_file, section),
            ]

    def __getitem__(self, item):
        for resolver in self.resolvers:
            value = resolver.get(item)
            if value:
                break
        else:
            logger.debug('Entry %s not found', item)
            raise KeyError()
        logger.debug('Entry %s has value %s', item, value)
        return value

    def get(self, item, default=None):
        try:
            return self[item]
        except KeyError:
            logger.debug('Key %s not found, returning default %s', item, default)
            return default



#!/usr/bin/env python
# -*- coding: utf-8 -*-
import os
import sys
from setuptools import setup


name = 's3conf'
package = 's3conf'
description = 'Utility package to help managing configuration files stored in S3-like services.'
url = 'https://github.com/sbneto/s3conf'
author = 'Samuel Martins Barbosa Neto'
author_email = 'samuel.m.b.neto@gmail.com'
license_ = 'MIT'


def get_packages(package):
    """
    Return root package and all sub-packages.
    """
    return [dirpath
            for dirpath, dirnames, filenames in os.walk(package)
            if os.path.exists(os.path.join(dirpath, '__init__.py'))]


def get_package_data(package):
    """
    Return all files under the root package, that are not in a
    package themselves.
    """
    walk = [(dirpath.replace(package + os.sep, '', 1), filenames)
            for dirpath, dirnames, filenames in os.walk(package)
            if not os.path.exists(os.path.join(dirpath, '__init__.py'))]

    filepaths = []
    for base, filenames in walk:
        filepaths.extend([os.path.join(base, filename)
                          for filename in filenames])
    return {package: filepaths}


with open('s3conf/version.py') as f:
    exec(f.read())
version = get_version(repo_path=os.path.dirname(__file__))

if sys.argv[-1] == 'publish':
    if os.system("pip freeze | grep wheel"):
        print("wheel not installed.\nUse `pip install wheel`.\nExiting.")
        sys.exit()
    os.system("python setup.py sdist upload")
    os.system("python setup.py bdist_wheel upload")
    print("You probably want to also tag the version now:")
    print("  git tag -a {0} -m 'version {0}'".format(version))
    print("  git push --tags")
    sys.exit()


setup(
    name=name,
    version=version,
    url=url,
    license=license_,
    description=description,
    long_description=open('README.md').read(),
    author=author,
    author_email=author_email,
    packages=get_packages(package),
    package_data=get_package_data(package),
    install_requires=[
        'boto3>=1.4.4',
        'click>=6.7',
        'python-editor>=1.0.3',
        'click-log>=0.2.1',
    ],
    python_requires='~=3.5',
    setup_requires=[],
    tests_require=['pytest', 'pytest-cov'],
    classifiers=[
        'Environment :: Web Environment',
        'Intended Audience :: Developers',
        'Programming Language :: Python :: 3 :: Only',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Topic :: Internet',
        'Topic :: Software Development :: Libraries :: Python Modules',
    ],
    entry_points='''
    [console_scripts]
    s3conf=s3conf.__main__:main
    ''',
)
